/* ///////////////////////////////////////////////////////////////////// */
/*!
 \file
 \brief Set particles initial conditions for the fluid-CR relative drift test.
 
 Initialize particles for the Fluid-particles relative drift test,
 section 4.3 of [MVBM18].
 
 \authors A. Mignone (mignone@ph.unito.it)\n
          B. Vaidya (bvaidya@unito.it)\n
 
 \date    March 20, 2018
  \b References: \n
   - [MVBM18] "A PARTICLE MODULE FOR THE PLUTO CODE: I - AN IMPLEMENTATION OF
               THE MHD-PIC EQUATIONS", Mignone etal.ApJS (2018)  [ Sec. 4.3 ]
*/
/* ///////////////////////////////////////////////////////////////////// */
#include "pluto.h"

/* ********************************************************************* */
void Particles_Init(Data *d, Grid *grid)
/*!
 *  Sets initial conditions on particles.
 *
 *  \param [in]    d       Pointer to the PLUTO data structure.
 *  \param [in]    grid    Pointer to the PLUTO grid structure.
 *
 *********************************************************************** */
{
  int i,j,k, np, dir;
  int np_glob = RuntimeGet()->Nparticles_glob;
  int np_cell = RuntimeGet()->Nparticles_cell;
  double xbeg[3], xend[3], dV;
  double gamma, c2 = PARTICLES_CR_C*PARTICLES_CR_C;
  Particle p;

/* ------------------------------------------------------------------
    Cell by cell initialization
    Note: do not use LoadRandom() to initialize spatial coordinates.
   ------------------------------------------------------------------ */

  if (np_cell > 0){
    int count = 0;
    DOM_LOOP(k,j,i){
      xbeg[IDIR] = grid->xl[IDIR][i]; xend[IDIR] = grid->xr[IDIR][i];
      xbeg[JDIR] = grid->xl[JDIR][j]; xend[JDIR] = grid->xr[JDIR][j];
      xbeg[KDIR] = grid->xl[KDIR][k]; xend[KDIR] = grid->xr[KDIR][k];

      dV = grid->dV[k][j][i];

  /* -- Loop on particles -- */
  
      for (np = 0; np < np_cell; np++){
        Particles_LoadUniform(np, np_cell, xbeg, xend, p.coord);

      /* -- Compute particle 3-vel -- */

        p.speed[IDIR] = g_inputParam[VPX1];
        p.speed[JDIR] = g_inputParam[VPX2];
        p.speed[KDIR] = 0.0;

      /* -- Assign particle 4-vel -- */

        gamma = DOT_PRODUCT(p.speed, p.speed);
        gamma = 1.0/sqrt(1.0 - gamma/c2);
        p.speed[IDIR] *= gamma;
        p.speed[JDIR] *= gamma;
        p.speed[KDIR] *= gamma;

        p.mass        = (1.e-2*g_inputParam[RHO_GAS]/np_cell)*dV;
        p.color       = 0.0;
        Particles_Insert (&p, d, PARTICLES_CREATE, grid);
        count++;
      }
    }
  } else {
    print ("! Particles_Init(): need np_cell > 0\n");
    QUIT_PLUTO(1);
  }

  Particles_SetID(d->PHead);
}

/* ********************************************************************* */
void Particles_Inject(Data *data, Grid *grid)
/*!
 *  Sets user-defined boundary conditions on particles.
 *
 *  \param [in]  data    Pointer to the PLUTO data structure.
 *  \param [in]  grid    Pointer to the PLUTO grid structure.
 *********************************************************************** */
{
}

/* ********************************************************************* */
void Particles_UserDefBoundary(Data *d, int side, Grid *grid)
/*
 *
 *********************************************************************** */
{
  int    dir;
  double xbeg[3], xend[3];
  particleNode *curr = d->PHead, *next;
  Particle *p;
  
  for (dir = 0; dir < 3; dir++) {
    xbeg[dir] = grid->xbeg_glob[dir];
    xend[dir] = grid->xend_glob[dir];
  }

  if (side == X1_BEG){
    dir = IDIR;
    while (curr != NULL){
      p    = &(curr->p);  
      next = curr->next; 
      if (p->coord[dir] < xbeg[dir]) Particles_Destroy (curr,d);
      curr = next;
    }
  }

  if (side == X1_END){
    dir = IDIR;
    while (curr != NULL){
      p    = &(curr->p);  
      next = curr->next; 
      if (p->coord[dir] > xend[dir]) Particles_Destroy (curr,d);
      curr = next;
    }
  }

  if (side == X2_BEG){
    dir = JDIR;
    while (curr != NULL){
      p    = &(curr->p);  
      next = curr->next; 
      if (p->coord[dir] < xbeg[dir]) Particles_Destroy (curr,d);
      curr = next;
    }
  }

  if (side == X2_END){
    dir = JDIR;
    while (curr != NULL){
      p    = &(curr->p);  
      next = curr->next; 
      if (p->coord[dir] > xend[dir]) Particles_Destroy (curr,d);
      curr = next;
    }
  }
  if (side == X3_BEG){
    dir = KDIR;
    while (curr != NULL){
      p    = &(curr->p);  
      next = curr->next; 
      if (p->coord[dir] < xbeg[dir]) Particles_Destroy (curr,d);
      curr = next;
    }
  }

  if (side == X3_END){
    dir = KDIR;
    while (curr != NULL){
      p    = &(curr->p);  
      next = curr->next; 
      if (p->coord[dir] > xend[dir]) Particles_Destroy (curr,d);
      curr = next;
    }
  }

}
