/* ///////////////////////////////////////////////////////////////////// */
/*! 
  \file  
  \brief Guided reconnection problem

  The init.c file collects most of the user-supplied functions useful 
  for problem configuration.
  It is automatically searched for by the makefile.

  \author A. Mignone (mignone@ph.unito.it)
  \date   Sepy 10, 2012
*/
/* ///////////////////////////////////////////////////////////////////// */
#include "pluto.h"

#define NMODES 40 
/* ********************************************************************* */
void Init (double *v, double x, double y, double z)
/*! 
 *
 *
 *********************************************************************** */
{
  static int first_call = 1;
  double a    = g_inputParam[DELTA];
  double pinf = sqrt(0.5*g_inputParam[BETA]);
  double cs;

  static double phim[NMODES];
  double Lx = g_domEnd[IDIR] - g_domBeg[IDIR];
  double k0 = (2.0*CONST_PI)/Lx;
  double dAz = 0.0;
  double k;

/* ----------------------------------------------
   0. Initialize prn sequence
   ---------------------------------------------- */

  if (first_call){
    RandomSeed (prank,0);
     for (int m = 0; m < NMODES; m++){
      phim[m] = RandomNumber(0.0,2.0*CONST_PI); /* Define random phases */
    }
    first_call = 0;
  }

/* ----------------------------------------------
   1. Impose equilibrium solution 
   ---------------------------------------------- */

  v[RHO] = 1.0;

  v[VX1] = 0.0;
  v[VX2] = 0.0;
  v[VX3] = 0.0;

  v[BX1] = tanh(y/a);  /* Mag. field defined such that vA = 1 */
  v[BX2] = 0.0;
  v[BX3] = 0.0;

  v[AX1] = 0.0;
  v[AX2] = 0.0;
  v[AX3] = a*log(cosh(y/a));

  v[PRS] = pinf - 0.5*v[BX1]*v[BX1] + 0.5;

/* ----------------------------------------------
   2. Perturbation
   ---------------------------------------------- */

  v[VX2] = -g_inputParam[VIN]*tanh(y/a)*exp(-pow(x/0.5,2.0));
  v[VX2] *= (1.0 + 0.1*RandomNumber(-1,1));



  double eps = 1.e-3;
  for (int m = 0; m < NMODES; m++) {
    k    = (m+1)*k0;
    dAz += sin(k*x + phim[m])*SECH(y/a)/k;
  }
  v[AX3] -= eps*dAz/NMODES;

}


/* ********************************************************************* */
void InitDomain (Data *d, Grid *grid)
/*! 
 * Assign initial condition by looping over the computational domain.
 * Called after the usual Init() function to assign initial conditions
 * on primitive variables.
 * Value assigned here will overwrite those prescribed during Init().
 *
 *
 *********************************************************************** */
{
}

/* ********************************************************************* */
void Analysis (const Data *d, Grid *grid)
/*! 
 *  Perform runtime data analysis.
 *
 * \param [in] d the PLUTO Data structure
 * \param [in] grid   pointer to array of Grid structures  
 *
 *********************************************************************** */
{

}
#if PHYSICS == MHD
/* ********************************************************************* */
void BackgroundField (double x1, double x2, double x3, double *B0)
/*!
 * Define the component of a static, curl-free background 
 * magnetic field.
 *
 * \param [in] x1  position in the 1st coordinate direction \f$x_1\f$
 * \param [in] x2  position in the 2nd coordinate direction \f$x_2\f$
 * \param [in] x3  position in the 3rd coordinate direction \f$x_3\f$
 * \param [out] B0 array containing the vector componens of the background
 *                 magnetic field
 *********************************************************************** */
{
   B0[0] = 0.0;
   B0[1] = 0.0;
   B0[2] = 0.0;
}
#endif

/* ********************************************************************* */
void UserDefBoundary (const Data *d, RBox *box, int side, Grid *grid) 
/*! 
 *
 *********************************************************************** */
{
  int   i, j, k, nv;
  double  *x1, *x2, *x3;
  double v[256];

  x1 = grid->x[IDIR];
  x2 = grid->x[JDIR];
  x3 = grid->x[KDIR];

  if (side == 0) {    /* -- check solution inside domain -- */
    DOM_LOOP(k,j,i){};
  }

  if (side == X2_BEG){  /* -- X2_BEG boundary -- */
    if (box->vpos == CENTER) {
      BOX_LOOP(box,k,j,i){  
        Init (v, x1[i], x2[j], x3[j]);
        NVAR_LOOP(nv) d->Vc[nv][k][j][i] = v[nv]; 
      }
    }else if (box->vpos == X1FACE){
      BOX_LOOP(box,k,j,i){  
        Init (v, x1[i], x2[j], x3[j]);
        d->Vs[BX1s][k][j][i] = v[BX1];
      }
    }else if (box->vpos == X2FACE){
      BOX_LOOP(box,k,j,i){  }
    }else if (box->vpos == X3FACE){
      BOX_LOOP(box,k,j,i){  }
    }
  }

  if (side == X2_END){  /* -- X2_END boundary -- */
    if (box->vpos == CENTER) {
      BOX_LOOP(box,k,j,i){  
        Init (v, x1[i], x2[j], x3[j]);
        NVAR_LOOP(nv) d->Vc[nv][k][j][i] = v[nv]; 
      }
    }else if (box->vpos == X1FACE){
      BOX_LOOP(box,k,j,i){  
        Init (v, x1[i], x2[j], x3[j]);
        d->Vs[BX1s][k][j][i] = v[BX1];
      }
    }else if (box->vpos == X2FACE){
      BOX_LOOP(box,k,j,i){  }
    }else if (box->vpos == X3FACE){
      BOX_LOOP(box,k,j,i){  }
    }
  }

}

#if BODY_FORCE != NO
/* ********************************************************************* */
void BodyForceVector(double *v, double *g, double x1, double x2, double x3)
/*!
 * Prescribe the acceleration vector as a function of the coordinates
 * and the vector of primitive variables *v.
 *
 * \param [in] v  pointer to a cell-centered vector of primitive 
 *                variables
 * \param [out] g acceleration vector
 * \param [in] x1  position in the 1st coordinate direction \f$x_1\f$
 * \param [in] x2  position in the 2nd coordinate direction \f$x_2\f$
 * \param [in] x3  position in the 3rd coordinate direction \f$x_3\f$
 *
 *********************************************************************** */
{
  g[IDIR] = 0.0;
  g[JDIR] = 0.0;
  g[KDIR] = 0.0;
}
/* ********************************************************************* */
double BodyForcePotential(double x1, double x2, double x3)
/*!
 * Return the gravitational potential as function of the coordinates.
 *
 * \param [in] x1  position in the 1st coordinate direction \f$x_1\f$
 * \param [in] x2  position in the 2nd coordinate direction \f$x_2\f$
 * \param [in] x3  position in the 3rd coordinate direction \f$x_3\f$
 * 
 * \return The body force potential \f$ \Phi(x_1,x_2,x_3) \f$.
 *
 *********************************************************************** */
{
  return 0.0;
}
#endif
